'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.Router = Router;

var _trimkit = require('trimkit');

var _utils = require('./utils');

/*
    Usage:

    Router(rootURL, routeSpec, onUnmatched):

    rootURL (string)
        the url prefix of all paths.

    routeSpec (object)

    onUnmatched (optional function(api, badPath, lastGoodRoute))
        Called whenever an unmatched route is navigated. Parameters are:
            api (Router instance)
            badpath (string)
            lastGoodRoute (object)
                a route descriptor object for the last good route.

    returns an api object:
        goto(path) - match the path to a route and navigate there
        href(pathName, vars) - build a path based on the name and supplied vars
        listen() - listen to window.onhashchange for route changes
        current() - get the current route descriptor object

    Example:

    const router = Router(
        '#',
        {
            'home': {
                path: '/',
                onenter: (r, route) => {
                    console.log('At application home');
                },
                onexit: (r, route, newRoute) => {
                    console.log(`Exiting from ${route.path} to ${newRoute.path}`);
                }
            },
            'article': {
                path: '/article/:id',
                vars: {id: /[a-f0-9]{6,40}/},
                onenter: (r, route) => {
                    console.log('Opening Article', route.vars.id);
                },
                onexit: (r, route, newRoute) => {
                    console.log('Closing Article', route.vars.id);
                }
            },
        },
        (r, path, lastGoodRoute) => {
            alert(`Unmatched route "${path}"`);
            r.goto(lastGoodRoute.path || 'home');
        }
    );
*/

function Router(baseUrl, routes, unmatched) {
    var listening = false;
    var currentRoute = _trimkit.Null;
    var reEnterHook = _trimkit.Null;

    var VARMATCH_RE = /:([^\/]+)/g;
    var ROUTEELEMENT_RE = /^[^\/]+$/;

    var routeMatcher = (0, _trimkit.ObjectKeys)(routes).map(function (name) {
        var route = routes[name];
        var reg = route.path.replace(VARMATCH_RE, function (m, varName) {
            var varDef = route.vars || {};
            var regExStr = '' + (varDef[varName] || /[^\/]+/);
            return '(' + regExStr.substring(1, regExStr.lastIndexOf('/')) + ')';
        });

        return _extends({
            matcher: new RegExp('^' + baseUrl + reg + '$'),
            name: name
        }, route);
    });

    function goto(urlOrName, vars) {
        var url = urlOrName.startsWith(baseUrl) ? urlOrName : href(urlOrName, vars);

        if (listening && _location() !== url) {
            // If we're not there make the change and exit.
            location.hash = url;
            return;
        }

        if (currentRoute && currentRoute.path === url) {
            // This is only supposed to happen when recovering from a bad path.
            return;
        }

        var routeVars = {};

        var routeMatch = routeMatcher.find(function (_ref) {
            var matcher = _ref.matcher,
                path = _ref.path;

            var match = url.match(matcher);
            if (!match) {
                return false;
            }

            if (path.indexOf(':') !== -1) {
                match.shift();
                // $FlowIgnore
                path.replace(VARMATCH_RE, function (_, varName) {
                    // We're abusing RegExp.replace here.
                    routeVars[varName] = match.shift();
                });
            }

            return true;
        });

        if (routeMatch) {
            (function () {
                var newRoute = {
                    name: routeMatch.name,
                    vars: routeVars,
                    path: url
                };
                if (currentRoute && currentRoute.name === newRoute.name && (0, _trimkit.isFunction)(reEnterHook)) {
                    // $FlowIgnore
                    reEnterHook(newRoute);
                    currentRoute = newRoute;
                } else {
                    var onexit = currentRoute && currentRoute.name ? routes[currentRoute.name].onexit || _utils.nop : _utils.nop;
                    Promise.resolve(onexit(api, currentRoute, newRoute)).then(function () {
                        reEnterHook = routes[routeMatch.name].onenter(api, newRoute);
                        currentRoute = newRoute;
                    });
                }
            })();
        } else if (unmatched) {
            unmatched(api, url, currentRoute);
        } else {
            if (currentRoute && listening) {
                location.hash = currentRoute.path;
                return;
            }
            throw new Error('No route for "' + url + '"');
        }
    };

    function href(routeName) {
        var vars = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

        var route = routes[routeName];
        if (!route) {
            throw new Error('Invalid route ' + routeName + '.');
        }

        var path = '' + route.path;

        if (route.vars) {
            path = path.replace(VARMATCH_RE, function (_, varName) {
                var value = vars[varName];
                if ((route.vars[varName] || ROUTEELEMENT_RE).test(value)) {
                    value = (0, _trimkit.isUndefined)(value) ? '' : '' + value;
                    return value;
                }

                throw new Error('Invalid value for route ' + path + ' var ' + varName + ': ' + value + '.');
            });
        }
        return '' + baseUrl + path;
    };

    function _location() {
        return location.hash;
    };

    function listen(initialRoute) {
        if (listening) {
            return;
        }

        self.addEventListener('hashchange', function () {
            return goto(_location());
        }, false);
        listening = true;
        goto(_location() || initialRoute);
    };

    function current() {
        return currentRoute;
    }

    var api = {
        goto: goto,
        href: href,
        listen: listen,
        current: current
    };

    return api;
}