// @flow

import { Component } from './component';
import { allocateNode, getNodeKey } from './node';
import { Renderer } from './renderer';
import { WorkerPortal } from 'worker-portal';

let rootCounter = 0;

export function mount<T>(
    componentOrWorker: ComponentType<T> | Worker,
    domRoot: ?Element,
    context: ?{[id: string]: any}
) {
    // Worker doesn't exist for Chrome inside WebWorker context so we check for it.
    if (self.Worker && componentOrWorker instanceof self.Worker) {
        return mountWorker(componentOrWorker, domRoot);
    }
    return mountComponent(componentOrWorker, domRoot, context);
}

export function mountComponent<T>(
    component: ComponentType<T>,
    domRoot: ?Element,
    context: ?{[id: string]: any}
) {
    const parent = allocateNode();
    const node = allocateNode();
    let instance;
    const rootId = rootCounter += 1;
    const rootContext = { ...context };

    node.component = component;
    parent.key = getNodeKey(parent);
    node.key = getNodeKey(node, parent, 0);

    if (!domRoot) {
        return new Promise(resolve => {
            let portal;

            WorkerPortal({
                mount: rootId => {
                    parent.id = rootId;
                    instance = Component(
                        node,
                        (patches) => {
                            portal.onpatch(patches);
                        },
                        parent,
                        {
                            ...rootContext,
                            ...portal,
                        }
                    );
                    resolve(instance);
                },
                dispatchEvent: evt => instance.dispatchEvent(evt),
                update: props => instance.update(props),
                remove: () => instance.remove(),
            }).then(p => {
                portal = p;
            });
        });
    } else {
        parent.id = rootId;

        function eventHandler(evt: any) {
            instance.dispatchEvent(evt);
        }

        instance = Component(
            node,
            Renderer(domRoot, rootId, eventHandler).queuePatch,
            parent,
            rootContext,
        );
        return instance;
    }
}

export function mountWorker(
    worker: Worker,
    domRoot: ?Element,
    context: ?{[id: string]: any}
) {
    let renderer;
    return WorkerPortal(
        {
            onpatch: (data) => {
                renderer.queuePatch(data);
            },
            ...context,
        },
        worker,
        true,
        (type, destination, id, params, defaultSerializer) => {
            try {
                return ('dataTransfer' in params[0]
                ? [type, destination, id, params]
                : defaultSerializer(type, destination, id, params));
            } catch (e) {
                return defaultSerializer(type, destination, id, params);
            }
        }
    ).then(portal => {
        const rootId = rootCounter += 1;

        renderer = Renderer(domRoot, rootId, portal.dispatchEvent);
        portal.mount(rootId);

        return portal;
    }).catch(e => {
        console.error(e.stack);
    });
}
