import {
    freeNode,
    convertTagsToNodes,
    shouldReplaceNode,
} from './node';
import { shallowEqual, isTag } from './utils';
import { isFunction, isUndefined, Null, ObjectKeys } from 'trimkit';
import {
    TEXT_NODE_VALUE,
    ADD_ACTION,
    UPDATE_ACTION,
    REMOVE_ACTION,
    REGISTER_STYLE,
} from './constants';

let eventCounter = 0;
const existingEventMap: Map<number, ElementRecordType> = new Map();
const events: Map<number, EventHandler> = new Map();
const selectorMatcher = /(?:(?:}|(?:(?:@[^{]+){))\s*)([^,@{]+)(?:,\s*([^,@{]+))*/gm;


export function Component<T>(
    node: StateTreeNodeType,
    onpatch: (patches: Array<PatchType>) => void,
    parent: StateTreeNodeType,
    parentContext: {[id:string]: any} = {}
): ComponentExternalApiType<T> {
    let state = [];
    let rawState = [];
    let pendingPatches = [];
    let checkWillUpdate = Null;
    let updateLock = false;
    let updateRequested = false;
    let instance;
    let previousProps;
    let previousChildren = [];
    let parentOffset = 0;
    let version = 0;
    // Clone the nodes so they can be freed in the parent's diff process.
    const hostNode = { ...node };
    const parentNode = { ...parent };
    const context = { ...parentContext };
    const componentMap = new Map();

    function dispatchEvent(data: PseudoEventType) {
        const callback = events.get(data.id);
        if (callback) {
            // Flow doesn't like my PseudoEventType going into a real EventHandler. Can't imagine why.
            // $FlowIgnore
            return callback(data);
        }
    }

    function replaceEventHandlers(_node: StateTreeNodeType) {
        const { name, attrs, id } = _node;
        for (let key in attrs) {
            const value = attrs[key];
            if (key.startsWith('on') || (key[0] === '_' && isFunction(value))) {
                if (!existingEventMap.has(id)) {
                    existingEventMap.set(id, {});
                }
                const elementRecord = existingEventMap.get(id);
                let eventId = elementRecord[key];

                if (isUndefined(eventId)) {
                    eventId = eventCounter++;
                    elementRecord[key] = eventId;
                    events.set(eventId, value);
                } else if (value !== events.get(eventId)){
                    events.set(eventId, value);
                }
                const flagKey = `_${key}_flags`;
                const flags = attrs[flagKey];
                attrs[key] = eventId;
                if (flags) {
                    attrs[flagKey] = flags.join(',');
                }
            }
        }

        return attrs;
    }

    function removeEventHandlers(elementId: number) {
        const elementRecord = existingEventMap.get(elementId);
        if (isUndefined(elementRecord)) {
            return;
        }

        for (let eventName in elementRecord) {
            events.delete(elementRecord[eventName]);
        }
        existingEventMap.delete(elementId);
    }

    function addUpdateChildComponent(parent, _node, indexOffset) {
        const { key } = _node;
        let childInstance = componentMap.get(key);
        if (!childInstance) {
            childInstance = Component(
                _node,
                onpatch,
                parent,
                context,
            );
            componentMap.set(key, childInstance);
        }
        const [patches, childRequestsUpdate, childCount] = childInstance.getPatches(
            _node.attrs,
            _node.children,
            indexOffset
        );
        updateRequested = updateRequested || childRequestsUpdate;
        return [patches, childCount];
    }

    function addNode(_node, parent, index) {
        if (_node.component) {
            return addUpdateChildComponent(parent, _node, index);
        }
        const { name, component } = _node;
        const patchAttrs = _node.attrs ? replaceEventHandlers(_node) : {};
        if (patchAttrs._willInsert) {
            updateLock = true;
            dispatchEvent({
                type: '_willInsert',
                id: patchAttrs._willInsert
            });
            updateLock = false;
        }
        return [
            [
                [
                    ADD_ACTION,
                    _node.id,
                    [component ? component.name : name, patchAttrs, parent.id, index]
                ]
            ].concat(_node.static ? [] : diff([], _node.children, _node, 0)),
            1
        ];
    }

    function removeNode(_node) {
        let patches;
        const processRemove = () => {
            removeEventHandlers(_node.id);
            return diff(
                _node.children,
                [],
                _node,
            ).concat([
                [REMOVE_ACTION, _node.id]
            ]);
        };

        if (_node.component) {
            const componentInstance = componentMap.get(_node.key);
            if (componentInstance) {
                patches = componentInstance.remove();
                componentMap.delete(_node.key);
            }
        } else if (_node.attrs._willRemove) {
            updateLock = true;
            const thennable = dispatchEvent({
                type: '_willRemove',
                id: _node.attrs._willRemove,
            });
            updateLock = false;
            if (thennable && isFunction(thennable.then)) {
                _node.delayFree = true;
                thennable.then(() => {
                    onpatch(processRemove());
                    _node.delayFree = false;
                    freeNode(_node);
                });
                patches = [];
            } else {
                patches = processRemove();
            }
        } else {
            patches = processRemove();
        }
        return patches;
    }

    function diff(currentState: StateTreeArrayType,
                  nextState: StateTreeArrayType,
                  parent: StateTreeNodeType,
                  indexOffset: number,
    ): Array<PatchType> {
        let patches: Array<PatchType> = [];
        const currentCount = currentState.length;
        const nextCount = nextState.length;

        let index = 0;

        while (index < currentCount) {
            let nextNode = index < nextCount ? nextState[index] : Null;
            const current = currentState[index];

            if (!nextNode) {
                patches.push.apply(patches, removeNode(current));
                if (current.delayFree) {
                    nextState[index] = current;
                    nextNode = current;
                }
            } else if (shouldReplaceNode(current, nextNode)) {
                patches.push.apply(
                    patches,
                    removeNode(current)
                );
                if (current.delayFree) {
                    nextState[index] = current;
                    freeNode(nextNode);
                    nextNode = current;
                } else {
                    const [nodePatches, childCount] = addNode(nextNode, parent, index + indexOffset);
                    patches.push.apply(patches, nodePatches);
                    index += childCount - 1;
                }
            } else if (nextNode.component) {
                // nodes are roughly equal but they're components so we compare them differently
                nextNode.key = current.key;
                const [nodePatches, childCount] = addUpdateChildComponent(parent, nextNode, index + indexOffset);
                patches.push.apply(patches, nodePatches);
                index += childCount - 1;
            } else {
                nextNode.id = current.id;

                const currentAttrs = current.attrs;
                const normalizedNextAttributes = nextNode.attrs ? replaceEventHandlers(nextNode) : {};
                const newAttrs = {};
                const rmAttrs = [];
                let attrsAreEqual = true;

                for (let key in currentAttrs) {
                    let val = normalizedNextAttributes[key];
                    if (!isUndefined(val)) {
                        if (val !== currentAttrs[key]) {
                            newAttrs[key] = val;
                            attrsAreEqual = false;
                        }
                    } else {
                        rmAttrs.push(key);
                        attrsAreEqual = false;
                    }
                }

                for (let key in normalizedNextAttributes) {
                    if (isUndefined(currentAttrs[key])) {
                        newAttrs[key] = normalizedNextAttributes[key];
                        attrsAreEqual = false;
                    }
                }

                if (!attrsAreEqual) {
                    // Update the current element
                    patches.push([
                        UPDATE_ACTION,
                        current.id,
                        newAttrs,
                        rmAttrs,
                    ]);
                }

                if (!nextNode.static) {
                    patches.push.apply(
                        patches,
                        diff(
                            current.children,
                            nextNode.children,
                            nextNode,
                            0,
                        )
                    );
                }
            }

            index++;
        }

        while (index < nextCount) {
            let nextNode = nextState[index];

            const [nodePatches, childCount] = addNode(nextNode, parent, index + indexOffset);
            patches.push.apply(patches, nodePatches);
            index += childCount; // - 1 + 1  -1 to correct for the count, +1 for adding this node.
        }

        return patches;
    }

    function updateChildContext(c) {
        ObjectKeys(c).forEach(k => context[k] = c[k]);
    }

    function registerStyle(css) {
        const key = getKey();
        hostNode.style.push(key);
        const outCSS = ('} ' + css.trim()).replace(selectorMatcher, (match, ...p) => {
            p.slice(0, p.length - 2).forEach(s => {
                if (s) {
                    match = match.split(s).join(`${s.trim()}[${key}]`);
                }
            });
            return match;
        }).slice(2);

        pendingPatches.push([
            REGISTER_STYLE,
            getKey(),
            outCSS,
        ]);
    }

    function getKey() {
        return hostNode.key;
    }

    function handleUpdate(checkFunction) {
        checkWillUpdate = checkFunction;
    }

    function update(props: any) {
        if (updateLock) {
            updateRequested = true;
            return;
        }

        const [patches, childRequestsUpdate] = getPatches(props);
        patches.length && onpatch(patches);
        if (childRequestsUpdate) {
            redraw();
        }
    }

    function _getPatches() {
        const newState = convertTagsToNodes(hostNode, rawState);
        const patches = diff(state, newState, parentNode, parentOffset);
        if (pendingPatches.length) {
            patches.push.apply(patches, pendingPatches);
            pendingPatches = [];
        }
        state.map(freeNode);
        state = newState;
        return patches;
    }

    function versionedRedrawFactory(frozenVersion) {
        return function versionedLockout() {
            if (frozenVersion === version) {
                redraw();
            }
        };
    }

    function getPatches(props: any, children: any, indexOffset: number) {
        if (isUndefined(props)) {
            props = previousProps;
        }

        if (isUndefined(children)) {
            children = previousChildren;
        } else {
            previousChildren = children;
        }

        if (!isUndefined(indexOffset)) {
            parentOffset = indexOffset;
        }

        updateLock = true;
        if (instance && checkWillUpdate && !checkWillUpdate(props, previousProps)) {
            previousProps = props;
        } else if (!instance || !shallowEqual(previousProps, props)) {
            version += 1;
            instance = hostNode.component({
                redraw: versionedRedrawFactory(version),
                getKey,
                registerStyle,
                handleUpdate,
                updateChildContext,
                ...context,
            }, props);
            previousProps = props;
        }
        updateLock = false;

        rawState = instance(children);
        const patches = _getPatches();
        const isRequestingUpdate = updateRequested;
        updateRequested = false;
        const childCount = isTag(rawState) ? 1 : rawState.length;
        return [patches, isRequestingUpdate, childCount];
    }

    function redraw() {
        if (!instance) {
            return;
        }
        update();
    }

    function remove() {
        version += 1;
        rawState = [];
        return _getPatches();
    }

    return {
        remove,
        redraw,
        update,
        getPatches,
        dispatchEvent,
    };
}
