// @flow

import {
    TEXT_NODE_VALUE,
    TAG_NAME,
    TAG_ATTRS,
    TAG_CHILDREN,
    GENERIC_ERROR_MESSAGE,
} from './constants';
import {
    isString,
    isNumber,
    isFunction,
    isUndefined,
    Null,
    isArray,
    ObjectKeys,
} from 'trimkit';
import {
    memoize,
    isTag,
    traverseThunk,
} from './utils';


const pool: StateTreeArrayType = [];
let nodeCounter = 0;

export function allocateNode(): StateTreeNodeType {
    const node = pool.length > 0 ? pool.pop() : {
        id: Null,
        name: '',
        attrs: {},
        children: [],
        static: false,
        svg: false,
        component: Null,
        key: '',
        style: [],
        delayFree: false
    };
    node.id = ++nodeCounter;
    return node;
}

export function freeNode(node: StateTreeNodeType) {
    if (node.delayFree) {
        return;
    }
    if (!node.component) {
        node.children.map(freeNode);
    }
    node.component = Null;
    node.static = node.svg = false;
    node.attrs = {};
    node.name = node.key = '';
    node.children = node.style = [];

    pool.push(node);
}

export function isObject(obj: any): bool {
    return (typeof obj === 'object') && (Null !== obj) && !isArray(obj);
}

export function isComponent(obj: TagInputType): bool {
    return isArray(obj) && isFunction(obj[TAG_NAME]);
}

export function shouldReplaceNode(currentNode, nextNode) {
    return nextNode.name !== currentNode.name
        || nextNode.component !== currentNode.component;
}

export function processTagClass(classAttr: any): Array<string> {
    let classObj = traverseThunk(classAttr);
    if (isObject(classObj)) {
        classObj = ObjectKeys(classObj).filter(key => traverseThunk(classObj[key]));
    } else if (isString(classObj)) {
        classObj = classObj.split(' ');
    }
    return classObj;
}

const attributesParser = /(?:(^|#|\.)([^#\.\[\]]+))|(\[(.+?)(?:\s*=\s*("|'|)((?:\\["'\]]|.)*?)\5)?\])/g;

export function parseTagAttrs(signature: string, attrs: AttributesType): [string, TagAttrsType] {
    let [tag, sigAttrs, classes] = parseTagSignature(signature);

    const newAttrs = {
        ...sigAttrs,
        class: classes.concat(processTagClass(attrs['class'] || ''))
            .concat(processTagClass(attrs.className || ''))
            .join(' ').trim()
    };

    ObjectKeys(attrs).forEach(key => {
        if (key.indexOf('class') === 0) {
            return;
        }
        const value = attrs[key];
        newAttrs[key] = (key.indexOf('on') === 0 || key[0] === '_') ? value : traverseThunk(value);
    });

    return [
        tag,
        newAttrs,
    ];
};

const parseTagSignature = memoize((signature: string) => {
    let tag = 'div';
    let classes = [];
    let match;
    const attrs = {};

    while ((match = attributesParser.exec(signature))) {
        if (match[1] === '' && match[2]) {
            tag = match[2];
        } else if (match[1] === '#') {
            attrs.id = match[2];
        } else if (match[1] === '.') {
            classes.push(match[2]);
        } else if (match[3][0] === '[') {
            var attrValue = match[6];
            if (attrValue) attrValue = attrValue.replace(/\\(["'])/g, '$1');
            attrs[match[4]] = attrValue || true;
        }
    }

    return [tag, attrs, classes];
});

export function getNodeKey(node: StateTreeNodeType, parent: ?StateTreeNodeType, index: ?number) {
    return node.attrs._key
        ? `${node.component.name || 'tag'}-${node.attrs._key}`
        : parent
        ? `sl_${parent.id.toString(36)}_${index.toString(36)}`
        : 'sl_root';
}

export function convertTagToNodes(
    parent: StateTreeNodeType,
    tag: TagType,
    index: number,
): StateTreeArrayType {
    // tag can be a thunk, string, array of tags or array of arrays.
    // we need to deal with each.

    // Thunk

    // FIXME - There could be an issue here where a thunk and a component could
    // be confused if the thunk is the first thing out of a render function.
    // Lock this down!
    if (isFunction(tag)) {
        const state = traverseThunk(tag, GENERIC_ERROR_MESSAGE);
        return convertTagsToNodes(parent, state);
    }

    let node = allocateNode();

    // String (Text tag)
    if (isString(tag) || isNumber(tag)) {
        node.name = TEXT_NODE_VALUE;
        node.attrs = {
            textContent: tag.toString(),
        };
        node.static = true;
    } else if (isTag(tag)) {
        // Element tag

        const hasAttrs = isObject(tag[TAG_ATTRS]);
        const attrs = hasAttrs ? tag[TAG_ATTRS] : {};

        [node.name, node.attrs] = parseTagAttrs(tag[TAG_NAME], attrs);
        node.key = node.attrs._key || node.attrs.id || Null;

        if (parent.svg || node.name === 'svg') {
            node.svg = true;
            node.attrs._ns = 'svg';
        }
        node.style = parent.style;
        node.style.forEach(s => { node.attrs[s] = ''; });  // Add style attributes to the tag.
        node.children = convertTagChildrenToNodes(
            node,
            tag.slice(hasAttrs ? TAG_CHILDREN : TAG_ATTRS)
        );
    } else if (isComponent(tag)) {
        const hasAttrs = isObject(tag[TAG_ATTRS]);
        const attrs = hasAttrs ? tag[TAG_ATTRS] : {};

        node.component = tag[TAG_NAME];
        node.attrs = attrs;
        node.key = getNodeKey(node, parent, index);
        node.style = parent.style.slice();
        node.static = true;
        node.children = tag.slice(hasAttrs ? TAG_CHILDREN : TAG_ATTRS);
    } else {
        freeNode(node);
        // Array
        return convertTagsToNodes(parent, tag);
    }

    return node;
}

function flatten(arr) {
    return arr.reduce((flat, toFlatten) => {
        return flat.concat(isArray(toFlatten) && !isTag(toFlatten) ? flatten(toFlatten) : toFlatten);
    }, []);
}

function convertTagChildrenToNodes(parent: StateTreeNodeType, tags: TagArrayType): StateTreeArrayType {
    if (isUndefined(tags) || Null === tags) {
        return [];
    }
    return flatten(tags.map((t, i) => convertTagToNodes(parent, t, i)));
}

export function convertTagsToNodes(parent: StateTreeNodeType, tags: TagInputType): StateTreeArrayType {
    if (isTag(tags) || isString(tags) || isComponent(tags)) {
        return [convertTagToNodes(parent, tags, 0)];
    }
    return convertTagChildrenToNodes(parent, tags);
}
