// @flow

import {
    PATCH_ELEMENT_ID,
    TEXT_NODE_VALUE,
    PATCH_DATA,
    PATCH_RM_DATA,
    PATCH_AM_DATA,
    PATCH_TYPE,
    ADD_ACTION,
    UPDATE_ACTION,
    REMOVE_ACTION,
    REGISTER_STYLE,
    REGISTER_STYLE_SELECTOR,
    REGISTER_STYLE_CSS,
    ELEMENT_PROPERTY,
    ELEMENT_ATTRIBUTE,
    ELEMENT_EVENT,
    ELEMENT_LIFECYCLE_EVENT,
    LOCAL_ELEMENT_LIFECYCLE_EVENT,
    LAYOUT_REQUEST_EVENT,
    RUN_AFTER_ACTION,
} from './constants';

import {
    requestAnimationFrame,
    isFunction,
    isUndefined,
    isArray,
    Null,
    DOMDocument,
    ObjectKeys,
} from 'trimkit';
import { getEventFlags } from './utils';

const RECTANGLE_PROPS = 'height width top right bottom left x y'.split(' ');

function injectElement(element: Element, parent: Element, index: number) {
    const nextSibling = parent.childNodes[index];
    if (nextSibling) {
        parent.insertBefore(element, nextSibling);
    } else {
        parent.appendChild(element);
    }
}

function callPropertyFactory(element, propName) {
    return () => element[propName]();
}

export function annotatePatchAttributes(patch) {
    if (patch[PATCH_TYPE] !== ADD_ACTION && patch[PATCH_TYPE] !== UPDATE_ACTION) {
        return patch;
    }

    const accessMap = {};
    const patchData = patch[PATCH_DATA];
    const patchAttrs = patch[PATCH_TYPE] === ADD_ACTION ? patchData[1] : patchData;
    for (let name in patchAttrs) {
        accessMap[name] = annotateAttribute(name);
    }

    if (patch[PATCH_TYPE] === UPDATE_ACTION) {
        patch[PATCH_RM_DATA].forEach((name) => {
            accessMap[name] = annotateAttribute(name);
        });
    }

    patch.push(accessMap);

    return patch;
}

const ACTION_ATTRIBUTES = ['focus','select','blur','click'];
const ELEMENT_PROPERTIES = 'textContent value checked disabled readOnly scrollTop scrollLeft'.split(' ');
const LOCAL_LIFECYCLE_PROPERTIES = ['_didInsert','_willInsert','_willRemove'];
const LIFECYCLE_PROPERTIES = '_didRemove';
const LAYOUT_REQUEST = '_onLayout';
export function annotateAttribute(name) {
    if (ELEMENT_PROPERTIES.indexOf(name) !== -1) {
        return ELEMENT_PROPERTY;
    } else if (LOCAL_LIFECYCLE_PROPERTIES.indexOf(name) !== -1) {
        return LOCAL_ELEMENT_LIFECYCLE_EVENT;
    } else if (LIFECYCLE_PROPERTIES === name) {
        return ELEMENT_LIFECYCLE_EVENT;
    } else if (LAYOUT_REQUEST === name) {
        return LAYOUT_REQUEST_EVENT;
    } else if (ACTION_ATTRIBUTES.indexOf(name) !== -1) {
        return RUN_AFTER_ACTION;
    } else if (name.indexOf('on') === 0) {
        return ELEMENT_EVENT;
    }
    return ELEMENT_ATTRIBUTE;
}

export function Renderer(domRoot: Element,
                         rootId: number = 0,
                         dispatchEvent: EventHandler) {
    const elementMap: Map<number, Element> = new Map();
    const eventMap: Map<number, Array<string> > = new Map();
    const styleMap: Map<string, string> = new Map();
    let pendingPatches: Array<Array<PatchType>> = [];
    let paused = false;
    const postProcessing = [];
    const layoutRequests = new Map();
    const layoutValues = new Map();

    window.addEventListener('resize', maintenance, false);

    elementMap.set(rootId, domRoot);

    function lifecycleEventFactory(type, id) {
        requestAnimationFrame(function() {
            dispatchEvent({
                type,
                id: parseInt(id, 10),
            });
        });
    }

    function eventHandler(evt: any) {
        if (!evt instanceof Event) {
            // Lifecycle events are not DOM events.
            dispatchEvent(evt);
        }
        const eventName = evt.type;
        const dataset = evt.currentTarget.dataset;
        const id = parseInt(dataset[`on${eventName}EventId`], 10);
        const flags = eventMap.get(id).split(',');
        const fakeEvt = {
            id,
            type: eventName,
        };
        let stop = true;
        let index = flags.length;
        while (index--) {
            const flag = flags[index];
            if (flag === 'bubble') {
                stop = false;
                continue;
            }
            const flagPaths = flag.split('.');
            const pathLastIndex = flagPaths.length - 1;
            let pointer = evt;
            let fakeEvtPointer = fakeEvt;
            flagPaths.forEach((attr, i) => {
                const newPointer = pointer[attr];
                if (i !== pathLastIndex) {
                    pointer = newPointer;
                    fakeEvtPointer = fakeEvtPointer[attr] || (fakeEvtPointer[attr] = {});
                } else if (isFunction(newPointer)) {
                    pointer[attr](); // run with context
                } else {
                    fakeEvtPointer[attr] = newPointer;
                }
            });
        }
        if (stop) {
            evt.stopPropagation();
        }
        dispatchEvent(fakeEvt);
    }

    function setAttributes(element: Element,
                           elementId: number,
                           attrs: AttributesType,
                           rmAttrs: Array<string>,
                           accessMap: AttributesType) {
        for (let name of ObjectKeys(attrs)) {
            const value = attrs[name];

            if (value === Null) {
                continue;
            }

            switch (accessMap[name]) {
            case ELEMENT_PROPERTY: {
                element[name] = value;
                continue;
            }
            case ELEMENT_EVENT: {
                const eventName = name.slice(2);
                const eventIdAttrName = `${name}EventId`;
                if (element.dataset[eventIdAttrName]) {
                    element.removeEventListener(eventName, eventHandler);
                }
                if (value) {
                    element.dataset[eventIdAttrName] = value;
                    eventMap.set(
                        value,
                        attrs[`_${name}_flags`] || getEventFlags(
                            eventName,
                            element.tagName.toLowerCase(),
                            attrs.type
                        )
                    );
                    element.addEventListener(eventName, eventHandler, false);
                }
                continue;
            }
            case ELEMENT_LIFECYCLE_EVENT: {
                element.dataset[name] = value;
                continue;
            }
            case ELEMENT_ATTRIBUTE: {
                element.setAttribute(name, value);
                continue;
            }
            case LAYOUT_REQUEST_EVENT: {
                layoutRequests.set(elementId, value);
                continue;
            }
            case RUN_AFTER_ACTION: {
                if (value) {
                    postProcessing.push(callPropertyFactory(element, name));
                }
            }
            default:
            }
        }

        for (let data of rmAttrs) {
            const [accessCode, name] = data;
            switch (accessCode) {
            case ELEMENT_EVENT:
                element.removeEventListener(eventName, eventHandler);
                continue;
            case ELEMENT_PROPERTY: {
                element[name] = value;
                continue;
            }
            case ELEMENT_ATTRIBUTE: {
                element.removeAttribute(name, value);
            }
            case LAYOUT_REQUEST_EVENT: {
                layoutRequests.delete(elementId);
                layoutValues.delete(elementId);
                continue;
            }

            case ELEMENT_LIFECYCLE_EVENT: {
                element.dataset.removeProp(name);
                continue;
            }
            default:
            }
        }
    }

    function addElement(patch: PatchAddType) {
        const elementId = patch[PATCH_ELEMENT_ID];
        const data = patch[PATCH_DATA];
        const accessMap = patch[PATCH_RM_DATA];
        const [ name, props, parentId, index ] = data;
        let element;

        if (name === TEXT_NODE_VALUE) {
            // $FlowIgnore
            element = DOMDocument.createTextNode(props.textContent);
        } else {
            if (props._ns === 'svg') {
                element = DOMDocument.createElementNS('http://www.w3.org/2000/svg', name);
            } else {
                element = DOMDocument.createElement(name);
                element.dataset.id = elementId;
            }

            setAttributes(element, elementId, props, [], accessMap);
        }
        elementMap.set(elementId, element);

        injectElement(element, elementMap.get(parentId), index);
        if (!isUndefined(props._didInsert)) {
            postProcessing.push(() => lifecycleEventFactory('_didInsert', props._didInsert));
        }
    }

    function updateElement(patch: PatchUpdateType) {
        const elementId = patch[PATCH_ELEMENT_ID];
        const updateAttrs = patch[PATCH_DATA];
        const rmAttrs = patch[PATCH_RM_DATA];
        const accessMap = patch[PATCH_AM_DATA];

        // $FlowIgnore
        setAttributes(elementMap.get(elementId), elementId, updateAttrs, rmAttrs, accessMap);
    }

    function removeElement(patch: PatchRemoveType) {
        const elementId = patch[PATCH_ELEMENT_ID];
        if (elementId && elementMap.has(elementId)) {
            // $FlowIgnore
            const element = elementMap.get(elementId);
            element.parentNode && element.parentNode.removeChild(element);
            layoutRequests.delete(elementId);
            layoutValues.delete(elementId);
            elementMap.delete(elementId);
            if (element.dataset && element.dataset._didRemove) {
                postProcessing.push(() => lifecycleEventFactory('_didRemove', element.dataset._didRemove));
            }
        }
    }

    function registerStyle(patch) {
        let selector = patch[1];
        let css = patch[2];
        let element = styleMap.get(selector);
        if (!element) {
            element = DOMDocument.createElement('style');
            styleMap.set(selector, element);
            DOMDocument.head.appendChild(element);
        }
        if (element.textContent !== css) {
            if (!css) {
                element.parentNode.removeChild(element);
                styleMap.delete(selector);
            } else {
                element.textContent = css;
            }
        }
    }

    function maintenance() {
        for (let key in styleMap.keys()) {
            if (!DOMDocument.querySelector('.' + key)) {
                const element = styleMap.get(key);
                styleMap.delete(key);
                element.parentNode.removeChild(element);
            }
        }

        for (let elementId of layoutRequests.keys()) {
            const id = layoutRequests.get(elementId);
            const rect = elementMap.get(elementId).getBoundingClientRect();
            const previousLayout = layoutValues.get(elementId) || {};

            if (RECTANGLE_PROPS.some(key => previousLayout[key] !== rect[key])) {
                const layoutEvt = {
                    id,
                    type: '_onLayout',
                };
                RECTANGLE_PROPS.forEach(key => layoutEvt[key] = rect[key]);

                layoutValues.set(elementId, layoutEvt);
                dispatchEvent(layoutEvt);
            };
        }
    }

    const ACTION_METHODS = {
        [ADD_ACTION]: addElement,
        [UPDATE_ACTION]: updateElement,
        [REMOVE_ACTION]: removeElement,
        // [MOVE_ACTION]: moveElement,
        [REGISTER_STYLE]: registerStyle,
    };

    const pause = () => {
        paused = true;
    };

    const resume = () => {
        paused = false;
        processPatch();
    };

    const queuePatch = (patchSet: Array<PatchType>, finish: bool = false) => {
        if (!patchSet) {
            return;
        }
        pendingPatches.unshift(patchSet.map(annotatePatchAttributes));
        if (finish) {
            processPatch();
            pause();
        } else {
            requestAnimationFrame(processPatch, domRoot);
        }
    };

    const processPatch = () => {
        if (paused) {
            return;
        }

        const patchSet = pendingPatches.pop();
        if (patchSet) {
            // console.group('PatchSet');
            let patch;
            while (patch = patchSet.shift()) {
                // console.log(JSON.stringify(patch));
                ACTION_METHODS[patch[PATCH_TYPE]](patch);
            }

            while(postProcessing.length) { postProcessing.pop()(); }
            // console.groupEnd('PatchSet');
        }
        if (pendingPatches.length) {
            requestAnimationFrame(processPatch, domRoot);
        } else {
            maintenance();
        }
    };

    requestAnimationFrame(processPatch, domRoot);

    return {
        queuePatch,
        pause,
        resume,
    };
}
