// @flow

import { TAG_NAME } from './constants';
import {
    Null,
    isFunction,
    isString,
    isUndefined,
    isArray,
    ObjectKeys,
    asArray,
} from 'trimkit';


export function nop() { return; };

export function traverseThunk(thunk: any, errorDefault: ?any) {
    if (!isFunction(thunk)) {
        return thunk;
    }
    let result = thunk;
    do {
        try {
            result = result();
        } catch (e) {
            if (isUndefined(errorDefault)) {
                throw e;
            }
            console.error(`Error rendering ${result.name}:`, e);
            result = errorDefault;
        }
    } while (isFunction(result));
    return result;
}

// From https://github.com/gaearon/react-pure-render/ (MIT)
export function shallowEqual<T>(objA: T, objB: T): bool {
    if (objA === objB) {
        return true;
    }

    if (
        typeof objA !== 'object'
        || Null === objA
        || typeof objB !== 'object'
        || Null === objB
    ) {
        return false;
    }

    var keysA = ObjectKeys(objA);
    var keysB = ObjectKeys(objB);

    if (keysA.length !== keysB.length) {
        return false;
    }

    // Test for A's keys different from B.
    var bHasOwnProperty = Object.prototype.hasOwnProperty.bind(objB);
    for (var i = 0; i < keysA.length; i++) {
        if (!bHasOwnProperty(keysA[i]) || objA[keysA[i]] !== objB[keysA[i]]) {
            return false;
        }
    }

    return true;
}


/*
* memoize.js
* by @philogb and @addyosmani
* with further optimizations by @mathias
* and @DmitryBaranovsk
* perf tests: http://bit.ly/q3zpG3
* Released under an MIT license.
*/

export function memoize(fn: Function) {
    fn.memoize = {};

    const wrappedFn = function() {
        var args = asArray(arguments),
            hash = '',
            i = args.length;
        let currentArg: any = Null;
        while (i--) {
            currentArg = args[i];
            hash += (currentArg === Object(currentArg))
                ? JSON.stringify(currentArg)
                : currentArg;
        }
        return (hash in fn.memoize)
            ? fn.memoize[hash]
            : fn.memoize[hash] = fn.apply(this, args);
    };

    wrappedFn.clear = function() {
        fn.memoize = {};
    };

    return wrappedFn;
}

export const getEventFlags = memoize(function getEventFlags(eventName, name, type) {
    const flags = [];

    if (name === 'input') {
        if (type === 'checkbox') {
            flags.push('target.checked');
        } else {
            flags.push('target.value');
        }
        if (['onkeyup', 'onkeydown', 'onkeypress'].indexOf(eventName) !== -1) {
            flags.push('keyCode', 'ctrlKey', 'altKey', 'shiftKey');
        }
    } else {
        flags.push('target.value');
    }
    return flags.join(',');
});

export function isTag(obj: TagInputType): bool {
    return isArray(obj) && isString(obj[TAG_NAME]);
}
